"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ElmToolingJsonManager = void 0;
const path_1 = __importDefault(require("path"));
const vscode_uri_1 = require("vscode-uri");
const util_1 = __importDefault(require("util"));
const fs = __importStar(require("fs"));
const readFile = util_1.default.promisify(fs.readFile);
class ElmToolingJsonManager {
    getEntrypoints(workspaceRootPath, sourceFile) {
        return __awaiter(this, void 0, void 0, function* () {
            const elmToolingPath = path_1.default.join(workspaceRootPath, "elm-tooling.json");
            const defaultRelativePathToFile = path_1.default.relative(workspaceRootPath, vscode_uri_1.URI.parse(sourceFile.uri).fsPath);
            return yield readFile(elmToolingPath, {
                encoding: "utf-8",
            })
                .then(JSON.parse)
                .then(this.elmToolingEntrypointsDecoder.bind(this))
                .then((entrypoints) => [
                entrypoints,
                `Using entrypoints from ${elmToolingPath}: ${JSON.stringify(entrypoints)}`,
            ], (error) => {
                const innerMessage = error.code === "ENOENT"
                    ? `No elm-tooling.json found in ${workspaceRootPath}.`
                    : error.code === "EISDIR"
                        ? `Skipping ${elmToolingPath} because it is a directory, not a file.`
                        : error instanceof SyntaxError
                            ? `Skipping ${elmToolingPath} because it contains invalid JSON: ${error.message}.`
                            : `Skipping ${elmToolingPath} because: ${error.message}.`;
                const fullMessage = `Using default entrypoint: ${defaultRelativePathToFile}. ${innerMessage}`;
                return [[defaultRelativePathToFile], fullMessage];
            });
        });
    }
    elmToolingEntrypointsDecoder(json) {
        if (typeof json !== "object" || json === null || Array.isArray(json)) {
            throw new Error(`Expected a JSON object but got: ${JSON.stringify(json)}`);
        }
        if (!("entrypoints" in json)) {
            throw new Error(`There is no "entrypoints" field.`);
        }
        const { entrypoints } = json;
        if (!Array.isArray(entrypoints) || entrypoints.length === 0) {
            throw new Error(`Expected "entrypoints" to be a non-empty array but got: ${JSON.stringify(json)}`);
        }
        const result = [];
        for (const [index, item] of entrypoints.entries()) {
            if (typeof item !== "string" || !item.startsWith("./")) {
                throw new Error(`Expected "entrypoints" to contain string paths starting with "./" but got: ${JSON.stringify(item)} at index ${index}`);
            }
            result.push(item);
        }
        return result;
    }
}
exports.ElmToolingJsonManager = ElmToolingJsonManager;
//# sourceMappingURL=elmToolingJsonManager.js.map