"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ElmMakeDiagnostics = exports.CODE_ACTION_ELM_MAKE = exports.NAMING_ERROR = void 0;
/* eslint-disable @typescript-eslint/no-unsafe-call */
const crypto_1 = require("crypto");
const path = __importStar(require("path"));
const tsyringe_1 = require("tsyringe");
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const utils = __importStar(require("../../compiler/utils/elmUtils"));
const elmWorkspaceMatcher_1 = require("../../util/elmWorkspaceMatcher");
const elmDiagnosticsHelper_1 = require("./elmDiagnosticsHelper");
const ELM_MAKE = "Elm";
exports.NAMING_ERROR = "NAMING ERROR";
const RANDOM_ID = crypto_1.randomBytes(16).toString("hex");
exports.CODE_ACTION_ELM_MAKE = `elmLS.elmMakeFixer-${RANDOM_ID}`;
class ElmMakeDiagnostics {
    constructor() {
        this.createDiagnostics = (sourceFile) => __awaiter(this, void 0, void 0, function* () {
            const filePath = vscode_uri_1.URI.parse(sourceFile.uri);
            const workspaceRootPath = this.elmWorkspaceMatcher
                .getProgramFor(filePath)
                .getRootPath();
            return yield this.checkForErrors(workspaceRootPath.fsPath, sourceFile).then((issues) => {
                return issues.length === 0
                    ? new Map([[filePath.toString(), []]])
                    : elmDiagnosticsHelper_1.ElmDiagnosticsHelper.issuesToDiagnosticMap(issues, workspaceRootPath);
            });
        });
        this.settings = tsyringe_1.container.resolve("Settings");
        this.connection = tsyringe_1.container.resolve("Connection");
        this.elmToolingJsonManager = tsyringe_1.container.resolve("ElmToolingJsonManager");
        this.elmWorkspaceMatcher = new elmWorkspaceMatcher_1.ElmWorkspaceMatcher((uri) => uri);
    }
    onCodeAction(params) {
        const { uri } = params.textDocument;
        const elmMakeDiagnostics = this.filterElmMakeDiagnostics(params.context.diagnostics);
        return this.convertDiagnosticsToCodeActions(elmMakeDiagnostics, uri);
    }
    hasType(error) {
        return "type" in error;
    }
    convertDiagnosticsToCodeActions(diagnostics, uri) {
        const result = [];
        diagnostics.forEach((diagnostic) => {
            if (diagnostic.message.startsWith(exports.NAMING_ERROR) ||
                diagnostic.message.startsWith("BAD IMPORT") ||
                diagnostic.message.startsWith("UNKNOWN LICENSE") ||
                diagnostic.message.startsWith("UNKNOWN PACKAGE") ||
                diagnostic.message.startsWith("UNKNOWN EXPORT")) {
                // Offer the name suggestions from elm make to our users
                const regex = /^\s{4}#(.*)#$/gm;
                let matches;
                while ((matches = regex.exec(diagnostic.message)) !== null) {
                    // This is necessary to avoid infinite loops with zero-width matches
                    if (matches.index === regex.lastIndex) {
                        regex.lastIndex++;
                    }
                    matches
                        .filter((_, groupIndex) => groupIndex === 1)
                        .forEach((match) => {
                        result.push(this.createQuickFix(uri, match, diagnostic, `Change to \`${match}\``));
                    });
                }
            }
            else if (diagnostic.message.startsWith("MODULE NAME MISMATCH") ||
                diagnostic.message.startsWith("UNEXPECTED SYMBOL")) {
                // Offer the name suggestions from elm make to our users
                const regex = /# -> #(.*)#$/gm;
                const matches = regex.exec(diagnostic.message);
                if (matches !== null) {
                    result.push(this.createQuickFix(uri, matches[1], diagnostic, `Change to \`${matches[1]}\``));
                }
            }
        });
        return result;
    }
    createQuickFix(uri, replaceWith, diagnostic, title) {
        const map = {};
        if (!map[uri]) {
            map[uri] = [];
        }
        map[uri].push(vscode_languageserver_1.TextEdit.replace(diagnostic.range, replaceWith));
        return {
            diagnostics: [diagnostic],
            edit: { changes: map },
            kind: vscode_languageserver_1.CodeActionKind.QuickFix,
            title,
        };
    }
    filterElmMakeDiagnostics(diagnostics) {
        return diagnostics.filter((diagnostic) => diagnostic.source === ELM_MAKE);
    }
    checkForErrors(workspaceRootPath, sourceFile) {
        return __awaiter(this, void 0, void 0, function* () {
            const settings = yield this.settings.getClientSettings();
            const [relativePathsToFiles, message] = yield this.elmToolingJsonManager.getEntrypoints(workspaceRootPath, sourceFile);
            this.connection.console.info(`Find entrypoints: ${message}. See https://github.com/elm-tooling/elm-language-server#configuration for more information.`);
            const argsMake = [
                "make",
                ...relativePathsToFiles,
                "--report",
                "json",
                "--output",
                "/dev/null",
            ];
            const argsTest = ["make", ...relativePathsToFiles, "--report", "json"];
            const makeCommand = settings.elmPath;
            const testCommand = settings.elmTestPath;
            const isTestFile = sourceFile.isTestFile;
            const args = isTestFile ? argsTest : argsMake;
            const testOrMakeCommand = isTestFile ? testCommand : makeCommand;
            const testOrMakeCommandWithOmittedSettings = isTestFile
                ? "elm-test"
                : "elm";
            const options = {
                cmdArguments: args,
                notFoundText: isTestFile
                    ? "'elm-test' is not available. Install Elm via 'npm install -g elm-test'."
                    : "The 'elm' compiler is not available. Install Elm via 'npm install -g elm'.",
            };
            try {
                // Do nothing on success, but return that there were no errors
                utils.execCmdSync(testOrMakeCommand, testOrMakeCommandWithOmittedSettings, options, workspaceRootPath, this.connection);
                return [];
            }
            catch (error) {
                if (typeof error === "string") {
                    return [];
                }
                else {
                    const execaError = error;
                    const lines = [];
                    execaError.stderr.split("\n").forEach((line) => {
                        let errorObject;
                        try {
                            errorObject = JSON.parse(line);
                        }
                        catch (error) {
                            this.connection.console.warn("Received an invalid json, skipping error.");
                        }
                        if (errorObject &&
                            this.hasType(errorObject) &&
                            errorObject.type === "compile-errors") {
                            const compilerError = errorObject;
                            compilerError.errors.forEach((error) => {
                                const problems = error.problems.map((problem) => ({
                                    details: problem.message
                                        .map((message) => typeof message === "string"
                                        ? message
                                        : `#${message.string}#`)
                                        .join(""),
                                    file: error.path
                                        ? path.isAbsolute(error.path)
                                            ? path.relative(workspaceRootPath, error.path)
                                            : error.path
                                        : relativePathsToFiles[0],
                                    overview: problem.title,
                                    region: problem.region,
                                    subregion: "",
                                    tag: "error",
                                    type: "error",
                                }));
                                lines.push(...problems);
                            });
                        }
                        else if (errorObject &&
                            this.hasType(errorObject) &&
                            errorObject.type === "error") {
                            const error = errorObject;
                            this.checkIfVersionMismatchesAndCreateMessage(error);
                            const problem = {
                                details: error.message
                                    .map((message) => typeof message === "string" ? message : message.string)
                                    .join(""),
                                // elm-test might supply absolute paths to files
                                file: error.path
                                    ? path.relative(workspaceRootPath, error.path)
                                    : relativePathsToFiles[0],
                                overview: error.title,
                                region: {
                                    end: {
                                        column: 1,
                                        line: 1,
                                    },
                                    start: {
                                        column: 1,
                                        line: 1,
                                    },
                                },
                                subregion: "",
                                tag: "error",
                                type: "error",
                            };
                            lines.push(problem);
                        }
                    });
                    return lines;
                }
            }
        });
    }
    checkIfVersionMismatchesAndCreateMessage(errorObject) {
        if (errorObject.title === "ELM VERSION MISMATCH") {
            this.connection.window.showErrorMessage(errorObject.message
                .map((message) => typeof message === "string" ? message : message.string)
                .join(""));
        }
    }
}
exports.ElmMakeDiagnostics = ElmMakeDiagnostics;
//# sourceMappingURL=elmMakeDiagnostics.js.map